package com.dsk.flink.dsc.common.sink;

import cn.hutool.core.lang.Snowflake;
import cn.hutool.core.util.IdUtil;
import cn.hutool.core.util.RandomUtil;
import cn.hutool.db.DbUtil;
import cn.hutool.db.sql.SqlExecutor;
import com.alibaba.druid.pool.DruidDataSource;
import com.dsk.flink.dsc.common.dto.SqlErrorLog;
import com.dsk.flink.dsc.utils.EnvProperties;
import org.apache.flink.api.java.tuple.Tuple6;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.sink.RichSinkFunction;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.util.Date;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

public class MysqlDataSlideSink extends RichSinkFunction<Tuple6<String,String,String,String,String,Long>> {

    static Logger logger = LoggerFactory.getLogger(MysqlDataSlideSink.class);
    EnvProperties envProps;
    private transient ExecutorService executorService;
    private transient DruidDataSource dataSource;
    private static final int MAX_RETRIES = 3; // 最大重试次数
    private static final int RETRY_DELAY_MS = 100; // 重试间隔时间
    private static final String SQL = "INSERT INTO dsc_cdc_log (`table`,op_type,pk_columns,pk_values,data_json,cdc_ts) values (?,?,?,?,?,?)";

    public MysqlDataSlideSink(EnvProperties envProps) {
        this.envProps = envProps;
    }

    @Override
    public void open(Configuration parameters) throws Exception {
        executorService = new ThreadPoolExecutor(5, 5, 20, TimeUnit.MINUTES, new LinkedBlockingDeque<>());
        //初始化获取配置
        String configTidbUrl = String.format(envProps.getDb_url(), envProps.getDb_host(), envProps.getDb_port(), envProps.getDb_database());
        //System.out.println(configTidbUrl);
        dataSource = new DruidDataSource();
        dataSource.setDriverClassName("com.mysql.cj.jdbc.Driver");
        dataSource.setUsername(envProps.getDb_username());
        dataSource.setPassword(envProps.getDb_password());
        dataSource.setUrl(configTidbUrl);
        dataSource.setMaxActive(30);
        dataSource.setInitialSize(20);
        dataSource.setTestWhileIdle(true);
        dataSource.setMaxWait(20000);
        dataSource.setValidationQuery("select 1");
    }

    @Override
    public void close() throws Exception {
        executorService.shutdown();
        dataSource.close();
    }

    @Override
    public void invoke(Tuple6<String,String,String,String,String,Long> value, Context context) throws Exception {
        executorService.execute(() -> {
            executeSqlWithRetry(value);
        });
    }

    private void executeSqlWithRetry(Tuple6<String,String,String,String,String,Long> t) {
        int retries = 0;
        Exception lastException = null;
        while (retries < MAX_RETRIES) {
            Connection connection = null;
            try {
                connection = dataSource.getConnection();
                SqlExecutor.execute(connection, SQL, t.f0,t.f1,t.f2,t.f3,t.f4,t.f5);
                return;
            } catch (Exception e) {
                lastException = e;
                retries++;
                logger.error("SQL执行失败，重试次数: {}, SQL: {}", retries, t.toString(), e);
                if (retries < MAX_RETRIES) {
                    try {
                        Thread.sleep((long) RETRY_DELAY_MS * retries);
                    } catch (InterruptedException ie) {
                        Thread.currentThread().interrupt();
                        logger.error("重试等待被中断", ie);
                    }
                }
            } finally {
                DbUtil.close(connection);
            }
        }
        // 所有重试都失败后记录错误日志
        logger.error("SQL执行失败，已达到最大重试次数: {}", MAX_RETRIES);
        SqlErrorLog errorLog = new SqlErrorLog(new Date(), t.toString(), lastException.getMessage());
        writeErrLogDb(errorLog);
    }

    private void writeErrLogDb(SqlErrorLog errorLog) {
        Snowflake snowflake = IdUtil.getSnowflake(RandomUtil.randomInt(31), RandomUtil.randomInt(31));
        String sql = "insert dsc_err_log (id,error_time, error_sql, error_msg) values (?, ?, ?, ?)";
        Connection conn = null;
        PreparedStatement pt = null;
        try {
            conn = dataSource.getConnection();
            pt = conn.prepareStatement(sql);
            pt.setLong(1, snowflake.nextId() + RandomUtil.randomInt(10,99));
            pt.setObject(2, errorLog.getErrorTime());
            pt.setString(3, errorLog.getSql());
            pt.setString(4, errorLog.getError());
            pt.execute();
        }catch (Exception e){
            logger.error("错误日志保存异常 -> {}", e.getMessage());
        }finally {
            DbUtil.close(pt, conn);
        }
    }
}
