package com.dsk.flink.dsc.sync;

import cn.hutool.core.map.MapUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson.JSONObject;
import com.dsk.flink.dsc.common.function.MysqlDataTransferFunction;
import com.dsk.flink.dsc.common.sink.MysqlDataSlideSink;
import com.dsk.flink.dsc.common.sink.MysqlDataTransferSink;
import com.dsk.flink.dsc.utils.EnvProperties;
import com.dsk.flink.dsc.utils.EnvPropertiesUtil;
import com.dsk.flink.dsc.utils.EtlUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.functions.FilterFunction;
import org.apache.flink.api.common.restartstrategy.RestartStrategies;
import org.apache.flink.api.common.serialization.SimpleStringSchema;
import org.apache.flink.api.java.tuple.Tuple3;
import org.apache.flink.api.java.tuple.Tuple6;
import org.apache.flink.api.java.utils.ParameterTool;
import org.apache.flink.streaming.api.CheckpointingMode;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.DataStreamSource;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.CheckpointConfig;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.assigners.TumblingProcessingTimeWindows;
import org.apache.flink.streaming.api.windowing.time.Time;
import org.apache.flink.streaming.api.windowing.windows.TimeWindow;
import org.apache.flink.streaming.connectors.kafka.FlinkKafkaConsumer;
import org.apache.flink.streaming.connectors.kafka.internals.KafkaTopicPartition;
import org.apache.flink.util.Collector;
import org.apache.flink.util.OutputTag;

import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Arrays;
import java.util.Map;

/**
 * @author shezaixing
 * @date 2023/12/5 14:44
 * @description 同步至客户目标数据源任务
 * update by lww
 */
@Slf4j
public class SyncCustomerDataSource {

    public static void main(String[] args) throws Exception {

        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();
        //env.setParallelism(1);
        env.setRestartStrategy(RestartStrategies.fixedDelayRestart(100, 60000));
        env.enableCheckpointing(120000);
        env.getCheckpointConfig().setCheckpointingMode(CheckpointingMode.EXACTLY_ONCE);
        env.getCheckpointConfig().setMinPauseBetweenCheckpoints(120000);
        env.getCheckpointConfig().setCheckpointTimeout(300000);
        env.getCheckpointConfig().setMaxConcurrentCheckpoints(1);
        env.getCheckpointConfig().setExternalizedCheckpointCleanup(CheckpointConfig.ExternalizedCheckpointCleanup.RETAIN_ON_CANCELLATION);
        env.getCheckpointConfig().setTolerableCheckpointFailureNumber(100);

        //获取用户自己的配置信息
        ParameterTool parameterTool = ParameterTool.fromArgs(args);
        String offsetTimestamp = parameterTool.get("offsetTimestamp");
        String position = parameterTool.get("position");

        String propertiesPath = parameterTool.get("propertiesPath");
        EnvProperties envProps = EnvPropertiesUtil.getPropertiesFromArgsPath(propertiesPath);
        System.out.println("读取到的配置文件：-> " + envProps.toString());
        System.out.println("读取到的数据连接配置：->" + String.format(envProps.getDb_url(), envProps.getDb_host(), envProps.getDb_port(), envProps.getDb_database()));
        System.out.println("获取到的kafka消费组：->" + EtlUtils.getKafkaGroup(envProps));
        FlinkKafkaConsumer<String> kafkaConsumer = new FlinkKafkaConsumer<String>(envProps.getKafka_topic(), new SimpleStringSchema(), EtlUtils.getKafkaConfig(envProps.getKafka_brokers(), EtlUtils.getKafkaGroup(envProps), envProps.getKafka_username(),envProps.getKafka_password()));
        long defaultOffset = LocalDateTime.now().minusMinutes(10).atZone(ZoneId.systemDefault()).toInstant().toEpochMilli();
        kafkaConsumer.setStartFromTimestamp(defaultOffset);
        //偏移量
        if (StrUtil.isNotBlank(offsetTimestamp)) {
            kafkaConsumer.setStartFromTimestamp(Long.parseLong(offsetTimestamp));
        }
        if (StrUtil.isNotBlank(position)) {
            // 0-1234,1-222
            Map<KafkaTopicPartition, Long> specificStartupOffsets = MapUtil.newHashMap();
            Arrays.stream(position.split(",")).forEach(x -> {
                String par = StrUtil.subBefore(x, "_", false);
                String pos = StrUtil.subAfter(x, "_", false);
                KafkaTopicPartition tp = new KafkaTopicPartition(envProps.getKafka_topic(),Integer.parseInt(par));
                specificStartupOffsets.put(tp,Long.parseLong(pos));
            });
            kafkaConsumer.setStartFromSpecificOffsets(specificStartupOffsets);
        }
        DataStreamSource<String> dscKafka = env
                .addSource(kafkaConsumer)
                .setParallelism(1);

        SingleOutputStreamOperator<JSONObject> tsGroupStream = dscKafka
                .map(JSONObject::parseObject)
                .setParallelism(1)
                .name("dsc-json")
                .uid("dsc-json")
                .filter(new FilterFunction<JSONObject>() {
                    @Override
                    public boolean filter(JSONObject value) throws Exception {
                        return !value.getBoolean("isDdl") && !"TIDB_WATERMARK".equals(value.getString("type"));
                    }
                })
                .setParallelism(1)
                .name("dsc-source")
                .uid("dsc-source");

        OutputTag<Tuple6<String,String,String,String,String,Long>> logSlideTag = new OutputTag<Tuple6<String,String,String,String,String,Long>>("log_slide") {};

        SingleOutputStreamOperator<Tuple3<String, String, Long>> slide = tsGroupStream
                .process(new MysqlDataTransferFunction(envProps,logSlideTag))
                .name("dsc-sql")
                .uid("dsc-sql");

        SingleOutputStreamOperator<String> groupWindowSqlResultStream = slide
                .keyBy(value -> value.f1)
                .window(TumblingProcessingTimeWindows.of(Time.milliseconds(500)))
                .process(new ProcessWindowFunction<Tuple3<String, String, Long>, String, String, TimeWindow>() {
                    @Override
                    public void process(String s, ProcessWindowFunction<Tuple3<String, String, Long>, String, String,
                            TimeWindow>.Context context, Iterable<Tuple3<String, String, Long>> elements,
                                        Collector<String> out) throws Exception {
                        Tuple3<String, String, Long> maxTsElement = null;
                        for (Tuple3<String, String, Long> element : elements) {
                            if (maxTsElement == null || element.f2 >= maxTsElement.f2) {
                                maxTsElement = element;
                            }
                        }
                        if (maxTsElement!= null) {
                            out.collect(maxTsElement.f0);
                        }
                    }
                })
                .name("dsc-max")
                .uid("dsc-max");

        groupWindowSqlResultStream.addSink(new MysqlDataTransferSink(envProps))
                .name("dsc-sink")
                .uid("dsc-sink");

        DataStream<Tuple6<String,String,String,String,String,Long>> sideOutput = slide.getSideOutput(logSlideTag);

        /*sideOutput.addSink(JdbcSink.sink(
                "INSERT INTO dsc_cdc_log (`table`,op_type,pk_columns,pk_values,data_json,cdc_ts) values (?,?,?,?,?,?)",
                (ps,t) -> {
                    ps.setString(1,t.f0);
                    ps.setString(2,t.f1);
                    ps.setString(3,t.f2);
                    ps.setString(4,t.f3);
                    ps.setString(5,t.f4);
                    ps.setLong(6,t.f5);
                },
                JdbcExecutionOptions.builder()
                        .withBatchSize(1000)
                        .withBatchIntervalMs(200)
                        .withMaxRetries(3)
                        .build(),
                new JdbcConnectionOptions.JdbcConnectionOptionsBuilder()
                        .withDriverName("com.mysql.cj.jdbc.Driver")
                        .withUrl(getSinkUrl(envProps))
                        .withUsername(envProps.getDb_username())
                        .withPassword(envProps.getDb_password())
                        .build()
        )).uid("dsc-log")
          .name("dsc-log");*/

        sideOutput.addSink(new MysqlDataSlideSink(envProps)).uid("dsc-log")
                .name("dsc-log");

        env.execute("dsc-client");
    }

    /*private static String getSinkUrl(EnvProperties envProps) {
        return String.format(envProps.getDb_url(), envProps.getDb_host(), envProps.getDb_port(), envProps.getDb_database());
    }*/
}
